#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fileio.h"
#include "parse_fat.h"

/*

This code is copyright 2007-2008 by Michael Kohn
You may not use any of this program in your own programs
without permission from Michael Kohn <mike@mikekohn.net>

*/

int parse_boot_sector(FILE *in, struct boot_sector_t *boot_sector)
{
  fread(boot_sector->jump_instr,1,3,in);
  read_chars(in,boot_sector->oem_name,8);
  boot_sector->bytes_per_sector=read_word(in);
  boot_sector->sectors_per_cluster=getc(in);
  boot_sector->reserved_sector_count=read_word(in);
  boot_sector->num_alloc_tables=getc(in);
  boot_sector->max_root_dir_entries=read_word(in);
  boot_sector->total_sectors=read_word(in);
  boot_sector->media_descriptor=getc(in);
  boot_sector->sectors_per_table=read_word(in);
  boot_sector->sectors_per_track=read_word(in);
  boot_sector->number_heads=read_word(in);
  boot_sector->hidden_sectors=read_int(in);
  boot_sector->total_sectors_4=read_int(in);

  /* Extended BIOS Param Block */

  boot_sector->physical_drive_num=getc(in);
  boot_sector->reserved=getc(in);
  boot_sector->signature=getc(in);
  read_chars(in,(char *)boot_sector->serial_num,4);
  read_chars(in,boot_sector->volume_label,11);
  read_chars(in,boot_sector->file_sys_type,8);
  fread(boot_sector->os_boot_code,1,448,in);
  fread(boot_sector->end_of_sector_marker,1,2,in);

  return 0;
}

static void print_media_descriptor(unsigned char d)
{
  if (d==0xf0)
  { printf("Double Sided, 80 Tracks, 18 Sectors/Track (1.44)\n"); }
    else
  if (d==0xf8)
  { printf("Single Sided, 80 Tracks, 9 Sectors/Track\n"); }
    else
  if (d==0xf9)
  { printf("Double Sided, 80 Tracks, 9 Sectors/Track\n"); }
    else
  if (d==0xfa)
  { printf("Single Sided, 80 Tracks, 8 Sectors/Track\n"); }
    else
  if (d==0xfb)
  { printf("Double Sided, 80 Tracks, 8 Sectors/Track\n"); }
    else
  if (d==0xfc)
  { printf("Single Sided, 40 Tracks, 9 Sectors/Track\n"); }
    else
  if (d==0xfd)
  { printf("Double Sided, 40 Tracks, 9 Sectors/Track\n"); }
    else
  if (d==0xfe)
  { printf("Single Sided, 40 Tracks, 8 Sectors/Track\n"); }
    else
  if (d==0xff)
  { printf("Single Sided, 40 Tracks, 8 Sectors/Track\n"); }

}

void print_boot_sector(struct boot_sector_t *boot_sector)
{
  printf("FAT Boot Sector\n");
  printf("-----------------------------------------------\n");

  printf("      Jump instruction: %02x %02x %02x\n",boot_sector->jump_instr[0],
                                                    boot_sector->jump_instr[1],
                                                    boot_sector->jump_instr[2]);
  printf("              OEM Name: %s\n",boot_sector->oem_name);
  printf("      Bytes Per Sector: %d\n",boot_sector->bytes_per_sector);

  printf("   Sectors Per Cluster: %d\n",boot_sector->sectors_per_cluster);
  printf(" Reserved Sector Count: %d\n",boot_sector->reserved_sector_count);
  printf("Number Of Alloc Tables: %d\n",boot_sector->num_alloc_tables);
  printf("  Max Root Dir Entries: %d\n",boot_sector->max_root_dir_entries);
  printf("         Total Sectors: %d\n",boot_sector->total_sectors);
  printf("       MediaDescriptor: (%02x) ",boot_sector->media_descriptor);

  print_media_descriptor(boot_sector->media_descriptor);

  printf("     Sectors Per Table: %d\n",boot_sector->sectors_per_table);
  printf("     Sectors Per Track: %d\n",boot_sector->sectors_per_track);
  printf("       Number Of Heads: %d\n",boot_sector->number_heads);
  printf("        Hidden Sectors: %d\n",boot_sector->hidden_sectors);
  printf("         Total Sectors: %d\n",boot_sector->total_sectors_4);

  printf("-- Extended BIOS Param Block --\n");
  printf(" Physical Drive Number: %d\n",boot_sector->physical_drive_num);
  printf("              Reserved: %d\n",boot_sector->reserved);
  printf("             Signature: %d\n",boot_sector->signature);
  printf("         Serial Number: %02x%02x%02x%02x\n",
                                        boot_sector->serial_num[0],
                                        boot_sector->serial_num[1],
                                        boot_sector->serial_num[2],
                                        boot_sector->serial_num[3]);
  printf("          Volume Label: '%s'\n",boot_sector->volume_label);
  printf("       Filesystem Type: '%s'\n",boot_sector->file_sys_type);
  printf("          OS Boot Code: (binary data)\n");
  printf("  End Of Sector Marker: %02x %02x\n",boot_sector->end_of_sector_marker[0],
                                               boot_sector->end_of_sector_marker[1]);
  printf("\n");

}

static void print_utf16_descramble(char *s, int len)
{
int chunks;
int n,t;

  chunks=len/26;

  for (t=0; t<chunks; t++)
  {
    for (n=len-26; n<len; n=n+2)
    {
      if (s[n]==0 && s[n+1]==0) break;
      printf("%c",s[n]);
    }

    len=len-26;
  }

}

static void print_time(int t, int d)
{
  printf("%02d:%02d:%02d %02d/%02d/%04d",
         (t>>11)&31,
         (t>>5)&63,
         (t&31)*2,
         ((d>>5)&15),
         (d&31),
         ((d>>9)&127)+1980);
}

static void print_dir_entry(struct dir_info_t *dir_info)
{
  printf("            Name: ");
  if ((unsigned char)dir_info->dos_name[0]==0x05)
  { printf("%c%s.%s\n",0xe5,dir_info->dos_name+1,dir_info->dos_ext); }
    else
  if ((unsigned char)dir_info->dos_name[0]==0xe5)
  { printf("?%s.%s (deleted)\n",dir_info->dos_name+1,dir_info->dos_ext); }
    else
  { printf("%s.%s\n",dir_info->dos_name,dir_info->dos_ext); }

  if (dir_info->vfat_name_len!=0)
  {
    printf("       VFAT Name: ");
    print_utf16_descramble((char *)dir_info->vfat_name,dir_info->vfat_name_len);
    printf("\n");
  }

  printf("      Attributes: ");
  if ((dir_info->attributes&1)==1) printf("[read-only]");
  if ((dir_info->attributes&2)==2) printf("[hidden]");
  if ((dir_info->attributes&4)==4) printf("[system]");
  if ((dir_info->attributes&8)==8) printf("[volume label]");
  if ((dir_info->attributes&16)==16) printf("[subdir]");
  if ((dir_info->attributes&32)==32) printf("[archive]");
  if ((dir_info->attributes&64)==64) printf("[device]");
  if ((dir_info->attributes&128)==128) printf("[unused]");
  printf("\n");

  printf("   Creation Time: ");
  print_time(dir_info->create_time, dir_info->create_date);
  printf("\n");
  printf("Last Access Time: ");
  print_time(dir_info->access_time,dir_info->access_date);
  printf("\n");
  printf("Last Modify Time: ");
  print_time(dir_info->modified_time,dir_info->modified_date);
  printf("\n");
  printf("   First Cluster: %d\n",dir_info->first_cluster);
  printf("       File Size: %d\n",dir_info->file_size);
  printf("\n");

}

static int parse_dir_entry(FILE *in, struct dir_info_t *dir_info)
{
  read_chars(in,dir_info->dos_name,8);
  read_chars(in,dir_info->dos_ext,3);
  dir_info->attributes=getc(in);
  dir_info->reserved=getc(in);
  dir_info->create_time_res=getc(in);

  if (dir_info->attributes!=0x0f)
  {
    dir_info->create_time=read_word(in);
    dir_info->create_date=read_word(in);
    dir_info->access_time=read_word(in);
    dir_info->access_date=read_word(in);
    dir_info->modified_time=read_word(in);
    dir_info->modified_date=read_word(in);
    dir_info->first_cluster=read_word(in);
    dir_info->file_size=read_int(in);
  }
    else
  {
    /* Read VFAT crap in */
#ifdef DEBUG
//printf("sequence number %d\n",dir_info->dos_name[0]);
#endif

    memcpy(dir_info->vfat_name+dir_info->vfat_name_len,dir_info->dos_name+1,7);
    memcpy(dir_info->vfat_name+dir_info->vfat_name_len+7,dir_info->dos_ext,3);
    read_chars(in,(char *)dir_info->vfat_name+dir_info->vfat_name_len+10,12);
    read_word(in);
    read_chars(in,(char *)dir_info->vfat_name+dir_info->vfat_name_len+22,4);
    dir_info->vfat_name_len+=26;

    parse_dir_entry(in,dir_info);
  }

  return 0;
}

int parse_dir(FILE *in, long offset, long partition_offset)
{
struct dir_info_t dir_info;

  fseek(in,offset+partition_offset,SEEK_SET);

  while(1)
  {
    memset(&dir_info,0,sizeof(struct dir_info_t));
    parse_dir_entry(in,&dir_info);
    if (dir_info.dos_name[0]==0) break;
    print_dir_entry(&dir_info);
  }

  return 0;
}




